/// Copyright (c) 2021 Razeware LLC
///
/// Permission is hereby granted, free of charge, to any person obtaining a copy
/// of this software and associated documentation files (the "Software"), to deal
/// in the Software without restriction, including without limitation the rights
/// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
/// copies of the Software, and to permit persons to whom the Software is
/// furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// Notwithstanding the foregoing, you may not use, copy, modify, merge, publish,
/// distribute, sublicense, create a derivative work, and/or sell copies of the
/// Software in any work that is designed, intended, or marketed for pedagogical or
/// instructional purposes related to programming, coding, application development,
/// or information technology.  Permission for such use, copying, modification,
/// merger, publication, distribution, sublicensing, creation of derivative works,
/// or sale is expressly withheld.
///
/// This project and source code may use libraries or frameworks that are
/// released under various Open-Source licenses. Use of those libraries and
/// frameworks are governed by their own individual licenses.
///
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
/// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
/// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
/// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
/// THE SOFTWARE.

import MetalKit

// swiftlint:disable collection_alignment
// swiftlint:disable comma
// swiftlint:disable indentation_width

typealias float3 = SIMD3<Float>
typealias float4 = SIMD4<Float>

struct Model {
  let meshes: [Mesh]
  let vertexDescriptor: MDLVertexDescriptor
  
  init(device: MTLDevice, name: String, vertexData: [Float]) {
    if let path = Bundle.main.path(forResource: "body-index", ofType: ".txt") {
      do {
        let data = try String(contentsOfFile: path, encoding: .utf8)
        let indicesData = Self.parseIndicesData(data)
        self.init(name: name, vertexData: vertexData, indicesData: indicesData)
      } catch {
        fatalError(error.localizedDescription)
      }
    } else {
      fatalError("body-index.txt not found!")
    }
  }
  
  init(name: String, vertexData: [Float], indicesData: [UInt32]) {
    let mdlMesh = Self.createMesh(device: Renderer.device, mesh: vertexData, indices: indicesData)
    vertexDescriptor = mdlMesh.vertexDescriptor
    print(vertexDescriptor)
    let mtkMesh = try! MTKMesh(mesh: mdlMesh, device: Renderer.device)
    
    meshes = [Mesh(mdlMesh: mdlMesh, mtkMesh: mtkMesh,
                   vertexDescriptor: vertexDescriptor)]
  }

  
  private static func parseIndicesData(_ data: String) -> [UInt32] {
    var indexData: [UInt32] = []
    let lines = data.components(separatedBy: "\r\n")
    for line in lines {
      if line.isEmpty {
          continue
      }
      let startLetter = line[0]
      if(startLetter == "g") {
        continue
      }
      let trimmedLine = line.trimmingCharacters(in: .whitespacesAndNewlines)
      let indices = getIndices(of: trimmedLine)
      indexData.append(contentsOf: indices)
    }
    
    
    
    return indexData
  }
  
  static func createMesh(device: MTLDevice, mesh rawVertexData: [Float], indices rawIndexData: [UInt32]) -> MDLMesh {
    let allocator = MTKMeshBufferAllocator(device: device)

    let mdlMeshVertexBuffer = allocator.newBuffer(MemoryLayout<Float>.stride * rawVertexData.count, type: .vertex)
    let vertexData = NSData(bytes: rawVertexData, length: MemoryLayout<Float>.stride * rawVertexData.count)
    print(rawVertexData.count)
    print(rawIndexData.count)
    mdlMeshVertexBuffer.fill(vertexData as Data, offset: 0)
    let indexBuffer = allocator.newBuffer(MemoryLayout<UInt32>.stride * rawIndexData.count, type: .index)
    let indexData = NSData(bytes: rawIndexData, length: MemoryLayout<UInt32>.stride * rawIndexData.count)
    indexBuffer.fill(indexData as Data, offset: 0)
//    let indexMap = indexBuffer.map()
//    indexMap.bytes.assumingMemoryBound(to: UInt32.self).assign(from: rawIndexData, count: rawIndexData.count)
    
    let material = MDLMaterial(name: "baseMaterial", scatteringFunction: MDLPhysicallyPlausibleScatteringFunction())
    let subMesh = MDLSubmesh(indexBuffer: indexBuffer, indexCount: rawIndexData.count, indexType: .uint32, geometryType: .triangles, material: material)
    
    let vertexDescriptor = MDLVertexDescriptor.defaultLayout
    let mdlMesh = MDLMesh(vertexBuffer: mdlMeshVertexBuffer, vertexCount: rawVertexData.count / 3, descriptor: vertexDescriptor, submeshes: [subMesh])
    mdlMesh.vertexDescriptor = vertexDescriptor
    return mdlMesh
  }

  private static func getIndices(of line: String) -> [UInt32] {
    let faceElements = line.split(separator: " ")
    let index0 = UInt32(faceElements[1].split(separator: "/")[0])! - 1
    let index1 = UInt32(faceElements[2].split(separator: "/")[0])! - 1
    let index2 = UInt32(faceElements[3].split(separator: "/")[0])! - 1
    let index3 = UInt32(faceElements[4].split(separator: "/")[0])! - 1
    
    return [index0, index1, index2, index0, index2, index3]
  }

}
